"""
pyart.retrieve.quasi_vertical_profile
=====================================

Retrieval of QVPs from a radar object

.. autosummary::
    :toctree: generated/

    quasi_vertical_profile
    compute_qvp
    compute_rqvp
    compute_evp
    compute_svp
    compute_vp
    compute_ts_along_coord
    compute_directional_stats
    project_to_vertical
    find_rng_index
    get_target_elevations
    find_nearest_gate
    find_neighbour_gates
    _create_qvp_object
    _create_along_coord_object
    _update_qvp_metadata
    _update_along_coord_metadata

"""

from copy import deepcopy
from warnings import warn

import numpy as np
from scipy.interpolate import interp1d
from netCDF4 import num2date

from ..core.transforms import antenna_to_cartesian
from ..io.common import make_time_unit_str
from ..util.xsect import cross_section_rhi, cross_section_ppi
from ..util.datetime_utils import datetime_from_radar


def quasi_vertical_profile(radar, desired_angle=None, fields=None, gatefilter=None):

    """
    Quasi Vertical Profile.

    Creates a QVP object containing fields from a radar object that can
    be used to plot and produce the quasi vertical profile

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field : string
        Radar field to use for QVP calculation.
    desired_angle : float
        Radar tilt angle to use for indexing radar field data.
        None will result in wanted_angle = 20.0

    Other Parameters
    ----------------
    gatefilter : GateFilter
        A GateFilter indicating radar gates that should be excluded
        from the import qvp calculation

    Returns
    -------
    qvp : Dictonary
        A quasi vertical profile object containing fields
        from a radar object

    References
    ----------
    Troemel, S., M. Kumjian, A. Ryzhkov, and C. Simmer, 2013: Backscatter
    differential phase - estimation and variability. J Appl. Meteor. Clim..
    52, 2529 - 2548.

    Troemel, S., A. Ryzhkov, P. Zhang, and C. Simmer, 2014: Investigations
    of backscatter differential phase in the melting layer. J. Appl. Meteorol.
    Clim. 54, 2344 - 2359.

    Ryzhkov, A., P. Zhang, H. Reeves, M. Kumjian, T. Tschallener, S. Tromel,
    C. Simmer, 2015: Quasi-vertical profiles - a new way to look at polarimetric
    radar data. Submitted to J. Atmos. Oceanic Technol.

    """

    # Creating an empty dictonary
    qvp = {}

    # Setting the desired radar angle and getting index value for desired radar angle
    if desired_angle is None:
        desired_angle = 20.0
    index = abs(radar.fixed_angle['data'] - desired_angle).argmin()
    radar_slice = radar.get_slice(index)

    # Printing radar tilt angles and radar elevation
    print(radar.fixed_angle['data'])
    print(radar.elevation['data'][-1])

    # Setting field parameters
    # If fields is None then all radar fields pulled else defined field is used
    if fields is None:
        fields = radar.fields

        for field in fields:

            # Filtering data based on defined gatefilter
            # If none is defined goes to else statement
            if gatefilter is not None:
                get_fields = radar.get_field(index, field)
                mask_fields = np.ma.masked_where(gatefilter.gate_excluded[radar_slice],
                                                 get_fields)
                radar_fields = np.ma.mean(mask_fields, axis=0)
            else:
                radar_fields = radar.get_field(index, field).mean(axis=0)

            qvp.update({field:radar_fields})

    else:
        # Filtering data based on defined gatefilter
        # If none is defined goes to else statement
        if gatefilter is not None:
            get_field = radar.get_field(index, fields)
            mask_field = np.ma.masked_where(gatefilter.gate_excluded[radar_slice],
                                            get_field)
            radar_field = np.ma.mean(mask_field, axis=0)
        else:
            radar_field = radar.get_field(index, fields).mean(axis=0)

        qvp.update({fields:radar_field})

    # Adding range, time, and height fields
    qvp.update({'range': radar.range['data'], 'time': radar.time})
    _, _, z = antenna_to_cartesian(qvp['range']/1000.0, 0.0,
                                   radar.fixed_angle['data'][index])
    qvp.update({'height': z})
    return qvp


def compute_qvp(radar, field_names, ref_time=None, angle=0., ang_tol=1.,
                hmax=10000., hres=50., avg_type='mean', nvalid_min=30,
                interp_kind='none', qvp=None):
    """
    Computes quasi vertical profiles.

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field_names : list of str
        list of field names to add to the QVP
    ref_time : datetime object
        reference time for current radar volume
    angle : int or float
        If the radar object contains a PPI volume, the sweep number to
        use, if it contains an RHI volume the elevation angle.
    ang_tol : float
        If the radar object contains an RHI volume, the tolerance in the
        elevation angle for the conversion into PPI
    hmax : float
        The maximum height to plot [m].
    hres : float
        The height resolution [m].
    avg_type : str
        The type of averaging to perform. Can be either "mean" or "median"
    nvalid_min : int
        Minimum number of valid points to accept average.
    interp_kind : str
        type of interpolation when projecting to vertical grid: 'none',
        or 'nearest', etc.
        'none' will select from all data points within the regular grid
        height bin the closest to the center of the bin.
        'nearest' will select the closest data point to the center of the
        height bin regardless if it is within the height bin or not.
        Data points can be masked values
        If another type of interpolation is selected masked values will be
        eliminated from the data points before the interpolation
    qvp : QVP object or None
        If it is not None this is the QVP object where to store the data from
        the current time step. Otherwise a new QVP object will be created


    Returns
    -------
    qvp : qvp object
        The computed QVP object

    Reference
    ---------
    Ryzhkov A., Zhang P., Reeves H., Kumjian M., Tschallener T., Trömel S.,
    Simmer C. 2016: Quasi-Vertical Profiles: A New Way to Look at Polarimetric
    Radar Data. JTECH vol. 33 pp 551-562

    """
    if avg_type not in ('mean', 'median'):
        warn('Unsuported statistics '+avg_type)
        return None

    radar_aux = deepcopy(radar)
    # transform radar into ppi over the required elevation
    if radar_aux.scan_type == 'rhi':
        radar_aux = cross_section_rhi(radar_aux, [angle], el_tol=ang_tol)
    elif radar_aux.scan_type == 'ppi':
        radar_aux = radar_aux.extract_sweeps([int(angle)])
    else:
        warn('Error: unsupported scan type.')
        return None

    if qvp is None:
        qvp = _create_qvp_object(
            radar_aux, field_names, qvp_type='qvp',
            start_time=ref_time, hmax=hmax, hres=hres)

    # modify metadata
    if ref_time is None:
        ref_time = datetime_from_radar(radar_aux)
    qvp = _update_qvp_metadata(
        qvp, ref_time, qvp.longitude['data'][0], qvp.latitude['data'][0],
        elev=qvp.fixed_angle['data'][0])

    for field_name in field_names:
        # compute QVP data
        if field_name not in radar_aux.fields:
            warn('Field '+field_name+' not in radar object')
            qvp_data = np.ma.masked_all(qvp.ngates)
        else:
            values, _ = compute_directional_stats(
                radar_aux.fields[field_name]['data'], avg_type=avg_type,
                nvalid_min=nvalid_min, axis=0)

            # Project to vertical grid:
            qvp_data = project_to_vertical(
                values, radar_aux.gate_altitude['data'][0, :],
                qvp.range['data'], interp_kind=interp_kind)

        # Put data in radar object
        if np.size(qvp.fields[field_name]['data']) == 0:
            qvp.fields[field_name]['data'] = qvp_data.reshape(1, qvp.ngates)
        else:
            qvp.fields[field_name]['data'] = np.ma.concatenate(
                (qvp.fields[field_name]['data'],
                 qvp_data.reshape(1, qvp.ngates)))

    return qvp


def compute_rqvp(radar, field_names, ref_time=None, hmax=10000., hres=2.,
                 avg_type='mean', nvalid_min=30, interp_kind='nearest',
                 rmax=50000., weight_power=2., qvp=None):
    """
    Computes range-defined quasi vertical profiles.

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field_names : list of str
        list of field names to add to the QVP
    ref_time : datetime object
        reference time for current radar volume
    hmax : float
        The maximum height to plot [m].
    hres : float
        The height resolution [m].
    avg_type : str
        The type of averaging to perform. Can be either "mean" or "median"
    nvalid_min : int
        Minimum number of valid points to accept average.
    interp_kind : str
        type of interpolation when projecting to vertical grid: 'none',
        or 'nearest', etc.
        'none' will select from all data points within the regular grid
        height bin the closest to the center of the bin.
        'nearest' will select the closest data point to the center of the
        height bin regardless if it is within the height bin or not.
        Data points can be masked values
        If another type of interpolation is selected masked values will be
        eliminated from the data points before the interpolation
    rmax : float
        ground range up to which the data is intended for use [m].
    weight_power : float
        Power p of the weighting function 1/abs(grng-(rmax-1))**p given to
        the data outside the desired range. -1 will set the weight to 0.
    qvp : QVP object or None
        If it is not None this is the QVP object where to store the data from
        the current time step. Otherwise a new QVP object will be created


    Returns
    -------
    qvp : qvp object
        The computed range defined quasi vertical profile

    Reference
    ---------
    Tobin D.M., Kumjian M.R. 2017: Polarimetric Radar and Surface-Based
    Precipitation-Type Observations of ice Pellet to Freezing Rain
    Transitions. Weather and Forecasting vol. 32 pp 2065-2082

    """
    if avg_type not in ('mean', 'median'):
        warn('Unsuported statistics '+avg_type)
        return None

    radar_aux = deepcopy(radar)
    # transform radar into ppi over the required elevation
    if radar_aux.scan_type == 'rhi':
        target_elevations, el_tol = get_target_elevations(radar_aux)
        radar_ppi = cross_section_rhi(
            radar_aux, target_elevations, el_tol=el_tol)
    elif radar_aux.scan_type == 'ppi':
        radar_ppi = radar_aux
    else:
        warn('Error: unsupported scan type.')
        return None

    if qvp is None:
        qvp = _create_qvp_object(
            radar_aux, field_names, qvp_type='rqvp',
            start_time=ref_time, hmax=hmax, hres=hres)

    # modify metadata
    if ref_time is None:
        ref_time = datetime_from_radar(radar_ppi)
    qvp = _update_qvp_metadata(
        qvp, ref_time, qvp.longitude['data'][0], qvp.latitude['data'][0],
        elev=90.)

    rmax_km = rmax/1000.
    grng_interp = np.ma.masked_all((radar_ppi.nsweeps, qvp.ngates))
    val_interp = dict()
    for field_name in field_names:
        val_interp.update({field_name: np.ma.masked_all(
            (radar_ppi.nsweeps, qvp.ngates))})

    for sweep in range(radar_ppi.nsweeps):
        radar_aux = deepcopy(radar_ppi)
        radar_aux = radar_aux.extract_sweeps([sweep])
        height = radar_aux.gate_altitude['data'][0, :]

        # compute ground range [Km]
        grng = np.sqrt(
            np.power(radar_aux.gate_x['data'][0, :], 2.) +
            np.power(radar_aux.gate_y['data'][0, :], 2.))/1000.

        # Project ground range to grid
        f = interp1d(
            height, grng, kind=interp_kind, bounds_error=False,
            fill_value='extrapolate')
        grng_interp[sweep, :] = f(qvp.range['data'])

        for field_name in field_names:
            if field_name not in radar_aux.fields:
                warn('Field '+field_name+' not in radar object')
                continue

            # Compute QVP for this sweep
            values, _ = compute_directional_stats(
                radar_aux.fields[field_name]['data'], avg_type=avg_type,
                nvalid_min=nvalid_min, axis=0)

            # Project to grid
            val_interp[field_name][sweep, :] = project_to_vertical(
                values, height, qvp.range['data'], interp_kind=interp_kind)

    # Compute weight
    weight = np.ma.abs(grng_interp-(rmax_km-1.))
    weight[grng_interp <= rmax_km-1.] = 1./np.power(
        weight[grng_interp <= rmax_km-1.], 0.)

    if weight_power == -1:
        weight[grng_interp > rmax_km-1.] = 0.
    else:
        weight[grng_interp > rmax_km-1.] = 1./np.power(
            weight[grng_interp > rmax_km-1.], weight_power)

    for field_name in field_names:

        # mask weights where there is no data
        mask = np.ma.getmaskarray(val_interp[field_name])
        weight_aux = np.ma.masked_where(mask, weight)

        # Weighted average
        qvp_data = (
            np.ma.sum(val_interp[field_name]*weight_aux, axis=0) /
            np.ma.sum(weight_aux, axis=0))

        # Put data in radar object
        if np.size(qvp.fields[field_name]['data']) == 0:
            qvp.fields[field_name]['data'] = qvp_data.reshape(1, qvp.ngates)
        else:
            qvp.fields[field_name]['data'] = np.ma.concatenate(
                (qvp.fields[field_name]['data'],
                 qvp_data.reshape(1, qvp.ngates)))

    return qvp


def compute_evp(radar, field_names, lon, lat, ref_time=None,
                latlon_tol=0.0005, delta_rng=15000., delta_azi=10,
                hmax=10000., hres=250., avg_type='mean', nvalid_min=1,
                interp_kind='none', qvp=None):
    """
    Computes enhanced vertical profiles.

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field_names : list of str
        list of field names to add to the QVP
    lat, lon : float
        latitude and longitude of the point of interest [deg]
    ref_time : datetime object
        reference time for current radar volume
    latlon_tol : float
        tolerance in latitude and longitude in deg.
    delta_rng, delta_azi : float
        maximum range distance [m] and azimuth distance [degree] from the
        central point of the evp containing data to average.
    hmax : float
        The maximum height to plot [m].
    hres : float
        The height resolution [m].
    avg_type : str
        The type of averaging to perform. Can be either "mean" or "median"
    nvalid_min : int
        Minimum number of valid points to accept average.
    interp_kind : str
        type of interpolation when projecting to vertical grid: 'none',
        or 'nearest', etc.
        'none' will select from all data points within the regular grid
        height bin the closest to the center of the bin.
        'nearest' will select the closest data point to the center of the
        height bin regardless if it is within the height bin or not.
        Data points can be masked values
        If another type of interpolation is selected masked values will be
        eliminated from the data points before the interpolation
    qvp : QVP object or None
        If it is not None this is the QVP object where to store the data from
        the current time step. Otherwise a new QVP object will be created


    Returns
    -------
    qvp : qvp object
        The computed enhanced vertical profile

    Reference
    ---------
    Kaltenboeck R., Ryzhkov A. 2016: A freezing rain storm explored with a
    C-band polarimetric weather radar using the QVP methodology. Meteorologische
    Zeitschrift vol. 26 pp 207-222

    """
    if avg_type not in ('mean', 'median'):
        warn('Unsuported statistics '+avg_type)
        return None

    radar_aux = deepcopy(radar)
    # transform radar into ppi over the required elevation
    if radar_aux.scan_type == 'rhi':
        target_elevations, el_tol = get_target_elevations(radar_aux)
        radar_ppi = cross_section_rhi(
            radar_aux, target_elevations, el_tol=el_tol)
    elif radar_aux.scan_type == 'ppi':
        radar_ppi = radar_aux
    else:
        warn('Error: unsupported scan type.')
        return None

    radar_aux = radar_ppi.extract_sweeps([0])

    if qvp is None:
        qvp = _create_qvp_object(
            radar_aux, field_names, qvp_type='evp',
            start_time=ref_time, hmax=hmax, hres=hres)

    # modify metadata
    if ref_time is None:
        ref_time = datetime_from_radar(radar_aux)
    qvp = _update_qvp_metadata(qvp, ref_time, lon, lat, elev=90.)

    height = dict()
    values = dict()
    for field_name in field_names:
        height.update({field_name: np.array([], dtype=float)})
        values.update({field_name: np.ma.array([], dtype=float)})

    for sweep in range(radar_ppi.nsweeps):
        radar_aux = deepcopy(radar_ppi)
        radar_aux = radar_aux.extract_sweeps([sweep])

        # find nearest gate to lat lon point
        ind_ray, _, azi, rng = find_nearest_gate(
            radar_aux, lat, lon, latlon_tol=latlon_tol)

        if ind_ray is None:
            continue

        # find neighbouring gates to be selected
        inds_ray, inds_rng = find_neighbour_gates(
            radar_aux, azi, rng, delta_azi=delta_azi, delta_rng=delta_rng)

        for field_name in field_names:
            if field_name not in radar_aux.fields:
                warn('Field '+field_name+' not in radar object')
                continue

            height[field_name] = np.append(
                height[field_name],
                radar_aux.gate_altitude['data'][ind_ray, inds_rng])

            # keep only data we are interested in
            field = radar_aux.fields[field_name]['data'][:, inds_rng]
            field = field[inds_ray, :]

            vals, _ = compute_directional_stats(
                field, avg_type=avg_type, nvalid_min=nvalid_min, axis=0)
            values[field_name] = np.ma.append(values[field_name], vals)

    for field_name in field_names:
        # Project to vertical grid:
        qvp_data = project_to_vertical(
            values[field_name], height[field_name], qvp.range['data'],
            interp_kind=interp_kind)

        # Put data in radar object
        if np.size(qvp.fields[field_name]['data']) == 0:
            qvp.fields[field_name]['data'] = qvp_data.reshape(1, qvp.ngates)
        else:
            qvp.fields[field_name]['data'] = np.ma.concatenate(
                (qvp.fields[field_name]['data'],
                 qvp_data.reshape(1, qvp.ngates)))

    return qvp


def compute_svp(radar, field_names, lon, lat, angle, ref_time=None,
                ang_tol=1., latlon_tol=0.0005, delta_rng=15000., delta_azi=10,
                hmax=10000., hres=250., avg_type='mean', nvalid_min=1,
                interp_kind='none', qvp=None):
    """
    Computes slanted vertical profiles.

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field_names : list of str
        list of field names to add to the QVP
    lat, lon : float
        latitude and longitude of the point of interest [deg]
    angle : int or float
        If the radar object contains a PPI volume, the sweep number to
        use, if it contains an RHI volume the elevation angle.
    ref_time : datetime object
        reference time for current radar volume
    ang_tol : float
        If the radar object contains an RHI volume, the tolerance in the
        elevation angle for the conversion into PPI
    latlon_tol : float
        tolerance in latitude and longitude in deg.
    delta_rng, delta_azi : float
        maximum range distance [m] and azimuth distance [degree] from the
        central point of the evp containing data to average.
    hmax : float
        The maximum height to plot [m].
    hres : float
        The height resolution [m].
    avg_type : str
        The type of averaging to perform. Can be either "mean" or "median"
    nvalid_min : int
        Minimum number of valid points to accept average.
    interp_kind : str
        type of interpolation when projecting to vertical grid: 'none',
        or 'nearest', etc.
        'none' will select from all data points within the regular grid
        height bin the closest to the center of the bin.
        'nearest' will select the closest data point to the center of the
        height bin regardless if it is within the height bin or not.
        Data points can be masked values
        If another type of interpolation is selected masked values will be
        eliminated from the data points before the interpolation
    qvp : QVP object or None
        If it is not None this is the QVP object where to store the data from
        the current time step. Otherwise a new QVP object will be created


    Returns
    -------
    qvp : qvp object
        The computed slanted vertical profile

    Reference
    ---------
    Bukovcic P., Zrnic D., Zhang G. 2017: Winter Precipitation Liquid-Ice
    Phase Transitions Revealed with Polarimetric Radar and 2DVD Observations
    in Central Oklahoma. JTECH vol. 56 pp 1345-1363

    """
    if avg_type not in ('mean', 'median'):
        warn('Unsuported statistics '+avg_type)
        return None

    radar_aux = deepcopy(radar)
    # transform radar into ppi over the required elevation
    if radar_aux.scan_type == 'rhi':
        radar_aux = cross_section_rhi(
            radar_aux, [angle], el_tol=ang_tol)
    elif radar_aux.scan_type == 'ppi':
        radar_aux = radar_aux.extract_sweeps([int(angle)])
    else:
        warn('Error: unsupported scan type.')
        return None

    if qvp is None:
        qvp = _create_qvp_object(
            radar_aux, field_names, qvp_type='svp',
            start_time=ref_time, hmax=hmax, hres=hres)

    # modify metadata
    if ref_time is None:
        ref_time = datetime_from_radar(radar_aux)
    qvp = _update_qvp_metadata(
        qvp, ref_time, lon, lat, elev=qvp.fixed_angle['data'][0])

    # find nearest gate to lat lon point
    ind_ray, _, azi, rng = find_nearest_gate(
        radar_aux, lat, lon, latlon_tol=latlon_tol)

    if ind_ray is None:
        warn('No data in selected point')
        qvp_data = np.ma.masked_all(qvp.ngates)

        for field_name in field_names:
            # Put data in radar object
            if np.size(qvp.fields[field_name]['data']) == 0:
                qvp.fields[field_name]['data'] = qvp_data.reshape(
                    1, qvp.ngates)
            else:
                qvp.fields[field_name]['data'] = np.ma.concatenate(
                    (qvp.fields[field_name]['data'],
                     qvp_data.reshape(1, qvp.ngates)))
        return qvp

    # find neighbouring gates to be selected
    inds_ray, inds_rng = find_neighbour_gates(
        radar_aux, azi, rng, delta_azi=delta_azi, delta_rng=delta_rng)
    height = radar_aux.gate_altitude['data'][ind_ray, inds_rng]

    for field_name in field_names:
        if field_name not in radar_aux.fields:
            warn('Field '+field_name+' not in radar object')
            qvp_data = np.ma.masked_all(qvp.ngates)
        else:
            # keep only data we are interested in
            field = radar_aux.fields[field_name]['data'][:, inds_rng]
            field = field[inds_ray, :]

            # compute values
            values, _ = compute_directional_stats(
                field, avg_type=avg_type, nvalid_min=nvalid_min, axis=0)

            # Project to vertical grid:
            qvp_data = project_to_vertical(
                values, height, qvp.range['data'], interp_kind=interp_kind)

        # Put data in radar object
        if np.size(qvp.fields[field_name]['data']) == 0:
            qvp.fields[field_name]['data'] = qvp_data.reshape(1, qvp.ngates)
        else:
            qvp.fields[field_name]['data'] = np.ma.concatenate(
                (qvp.fields[field_name]['data'],
                 qvp_data.reshape(1, qvp.ngates)))

    return qvp


def compute_vp(radar, field_names, lon, lat, ref_time=None,
               latlon_tol=0.0005, hmax=10000., hres=50.,
               interp_kind='none', qvp=None):
    """
    Computes vertical profiles.

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field_names : list of str
        list of field names to add to the QVP
    lat, lon : float
        latitude and longitude of the point of interest [deg]
    ref_time : datetime object
        reference time for current radar volume
    latlon_tol : float
        tolerance in latitude and longitude in deg.
    hmax : float
        The maximum height to plot [m].
    hres : float
        The height resolution [m].
    interp_kind : str
        type of interpolation when projecting to vertical grid: 'none',
        or 'nearest', etc.
        'none' will select from all data points within the regular grid
        height bin the closest to the center of the bin.
        'nearest' will select the closest data point to the center of the
        height bin regardless if it is within the height bin or not.
        Data points can be masked values
        If another type of interpolation is selected masked values will be
        eliminated from the data points before the interpolation
    qvp : QVP object or None
        If it is not None this is the QVP object where to store the data from
        the current time step. Otherwise a new QVP object will be created


    Returns
    -------
    qvp : qvp object
        The computed vertical profile

    """
    radar_aux = deepcopy(radar)
    # transform radar into ppi over the required elevation
    if radar_aux.scan_type == 'rhi':
        target_elevations, el_tol = get_target_elevations(radar_aux)
        radar_ppi = cross_section_rhi(
            radar_aux, target_elevations, el_tol=el_tol)
    elif radar_aux.scan_type == 'ppi':
        radar_ppi = radar_aux
    else:
        warn('Error: unsupported scan type.')
        return None

    if qvp is None:
        qvp = _create_qvp_object(
            radar_ppi, field_names, qvp_type='vp',
            start_time=ref_time, hmax=hmax, hres=hres)

    # modify metadata
    if ref_time is None:
        ref_time = datetime_from_radar(radar_ppi)
    qvp = _update_qvp_metadata(qvp, ref_time, lon, lat, elev=90.)

    height = dict()
    values = dict()
    for field_name in field_names:
        height.update({field_name: np.array([], dtype=float)})
        values.update({field_name: np.ma.array([], dtype=float)})

    for sweep in range(radar_ppi.nsweeps):
        radar_aux = deepcopy(radar_ppi.extract_sweeps([sweep]))

        # find nearest gate to lat lon point
        ind_ray, ind_rng, _, _ = find_nearest_gate(
            radar_aux, lat, lon, latlon_tol=latlon_tol)

        if ind_ray is None:
            continue

        for field_name in field_names:
            if field_name not in radar_aux.fields:
                warn('Field '+field_name+' not in radar object')
                continue

            height[field_name] = np.append(
                height[field_name],
                radar_aux.gate_altitude['data'][ind_ray, ind_rng])
            values[field_name] = np.ma.append(
                values[field_name],
                radar_aux.fields[field_name]['data'][ind_ray, ind_rng])

    for field_name in field_names:
        # Project to vertical grid:
        qvp_data = project_to_vertical(
            values[field_name], height[field_name], qvp.range['data'],
            interp_kind=interp_kind)

        # Put data in radar object
        if np.size(qvp.fields[field_name]['data']) == 0:
            qvp.fields[field_name]['data'] = qvp_data.reshape(1, qvp.ngates)
        else:
            qvp.fields[field_name]['data'] = np.ma.concatenate(
                (qvp.fields[field_name]['data'],
                 qvp_data.reshape(1, qvp.ngates)))

    return qvp


def compute_ts_along_coord(radar, field_name, mode='ALONG_AZI',
                           fixed_range=None, fixed_azimuth=None,
                           fixed_elevation=None, ang_tol=1., rng_tol=50.,
                           value_start=None, value_stop=None, ref_time=None,
                           acoord=None):
    """
    Computes time series along a particular antenna coordinate, i.e. along
    azimuth, elevation or range

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field_name : str
        Name of the field
    mode : str
        coordinate to extract data along. Can be ALONG_AZI, ALONG_ELE or
        ALONG_RNG
    fixed_range, fixed_azimuth, fixed_elevation : float
        The fixed range [m], azimuth [deg] or elevation [deg] to extract.
        In each mode two of these parameters have to be defined. If they are
        not defined they default to 0.
    ang_tol, rng_tol : float
        The angle tolerance [deg] and range tolerance [m] around the fixed
        range or azimuth/elevation
    value_start, value_stop : float
        The minimum and maximum value at which the data along a coordinate
        start and stop
    ref_time : datetime object
        reference time for current radar volume
    acoord : acoord object or None
        If it is not None this is the object where to store the data from the
        current time step. Otherwise a new acoord object will be created

    Returns
    -------
    acoord : acoord object
        The computed data along a coordinate

    """
    if mode == 'ALONG_RNG':
        if value_start is None:
            value_start = 0.
        if value_stop is None:
            value_stop = radar.range['data'][-1]
        if fixed_azimuth is None:
            fixed_azimuth = 0.
        if fixed_elevation is None:
            fixed_elevation = 0.
    elif mode == 'ALONG_AZI':
        if value_start is None:
            value_start = np.min(radar.azimuth['data'])
        if value_stop is None:
            value_stop = np.max(radar.azimuth['data'])
        if fixed_range is None:
            fixed_range = 0.
        if fixed_elevation is None:
            fixed_elevation = 0.
    elif mode == 'ALONG_ELE':
        if value_start is None:
            value_start = np.min(radar.elevation['data'])
        if value_stop is None:
            value_stop = np.max(radar.elevation['data'])
        if fixed_range is None:
            fixed_range = 0.
        if fixed_azimuth is None:
            fixed_azimuth = 0.
    else:
        warn('Unknown time series of coordinate mode '+mode)
        return None

    if mode == 'ALONG_RNG':
        # rng_values : range
        # fixed_angle: elevation
        # elevation: elevation
        # azimuth: azimuth
        rng_values, vals, _, _ = get_data_along_rng(
            radar, field_name, [fixed_elevation], [fixed_azimuth],
            ang_tol=ang_tol, rmin=value_start, rmax=value_stop)

        if vals.size == 0:
            warn('No data found')
            return None
        fixed_angle = fixed_elevation
        elevation = fixed_elevation
        azimuth = fixed_azimuth
        atol = rng_tol
    elif mode == 'ALONG_AZI':
        # rng_values : azimuth
        # fixed_angle : elevation
        # elevation : elevation
        # azimuth : range
        rng_values, vals, _, _ = get_data_along_azi(
            radar, field_name, [fixed_range], [fixed_elevation],
            rng_tol=rng_tol, ang_tol=ang_tol, azi_start=value_start,
            azi_stop=value_stop)

        if vals.size == 0:
            warn('No data found')
            return None
        fixed_angle = fixed_elevation
        elevation = fixed_elevation
        azimuth = fixed_range
        atol = ang_tol
    else:
        # rng_values : elevation
        # fixed_angle : azimuth
        # elevation : range
        # azimuth : azimuth
        rng_values, vals, _, _ = get_data_along_ele(
            radar, field_name, [fixed_range], [fixed_azimuth],
            rng_tol=rng_tol, ang_tol=ang_tol, ele_min=value_start,
            ele_max=value_stop)

        if vals.size == 0:
            warn('No data found')
            return None
        fixed_angle = fixed_azimuth
        elevation = fixed_range
        azimuth = fixed_azimuth
        atol = ang_tol

    if acoord is None:
        acoord = _create_along_coord_object(
            radar, [field_name], rng_values, fixed_angle, mode,
            start_time=ref_time)

    if not np.allclose(rng_values, acoord.range['data'], rtol=1e5, atol=atol):
        warn('Unable to add data. xvalues different from previous ones')
        return None

    # modify metadata
    if ref_time is None:
        ref_time = datetime_from_radar(radar)
    acoord = _update_along_coord_metadata(
        acoord, ref_time, elevation, azimuth)

    # Put data in radar object
    if np.size(acoord.fields[field_name]['data']) == 0:
        acoord.fields[field_name]['data'] = vals.reshape(1, acoord.ngates)
    else:
        acoord.fields[field_name]['data'] = np.ma.concatenate(
            (acoord.fields[field_name]['data'],
             vals.reshape(1, acoord.ngates)))

    return acoord


def compute_directional_stats(field, avg_type='mean', nvalid_min=1, axis=0):
    """
    Computes the mean or the median along one of the axis (ray or range)

    Parameters
    ----------
    field : ndarray
        the radar field
    avg_type :str
        the type of average: 'mean' or 'median'
    nvalid_min : int
        the minimum number of points to consider the stats valid. Default 1
    axis : int
        the axis along which to compute (0=ray, 1=range)

    Returns
    -------
    values : ndarray 1D
        The resultant statistics
    nvalid : ndarray 1D
        The number of valid points used in the computation

    """
    if avg_type == 'mean':
        values = np.ma.mean(field, axis=axis)
    else:
        values = np.ma.median(field, axis=axis)

    # Set to non-valid if there is not a minimum number of valid gates
    valid = np.logical_not(np.ma.getmaskarray(field))
    nvalid = np.sum(valid, axis=0, dtype=int)
    values[nvalid < nvalid_min] = np.ma.masked

    return values, nvalid


def project_to_vertical(data_in, data_height, grid_height, interp_kind='none',
                        fill_value=-9999.):
    """
    Projects radar data to a regular vertical grid

    Parameters
    ----------
    data_in : ndarray 1D
        the radar data to project
    data_height : ndarray 1D
        the height of each radar point
    grid_height : ndarray 1D
        the regular vertical grid to project to
    interp_kind : str
        The type of interpolation to use: 'none' or 'nearest'
    fill_value : float
        The fill value used for interpolation

    Returns
    -------
    data_out : ndarray 1D
        The projected data

    """
    if data_in.size == 0:
        data_out = np.ma.masked_all(grid_height.size)
        return data_out

    if interp_kind == 'none':
        hres = grid_height[1]-grid_height[0]
        data_out = np.ma.masked_all(grid_height.size)
        for ind_r, h in enumerate(grid_height):
            ind_h = find_rng_index(data_height, h, rng_tol=hres/2.)
            if ind_h is None:
                continue
            data_out[ind_r] = data_in[ind_h]
    elif interp_kind == 'nearest':
        data_filled = data_in.filled(fill_value=fill_value)
        f = interp1d(
            data_height, data_filled, kind=interp_kind, bounds_error=False,
            fill_value=fill_value)
        data_out = np.ma.masked_values(f(grid_height), fill_value)
    else:
        valid = np.logical_not(np.ma.getmaskarray(data_in))
        height_valid = data_height[valid]
        data_valid = data_in[valid]
        f = interp1d(
            height_valid, data_valid, kind=interp_kind, bounds_error=False,
            fill_value=fill_value)
        data_out = np.ma.masked_values(f(grid_height), fill_value)

    return data_out


def find_rng_index(rng_vec, rng, rng_tol=0.):
    """
    Find the range index corresponding to a particular range

    Parameters
    ----------
    rng_vec : float array
        The range data array where to look for
    rng : float
        The range to search
    rng_tol : float
        Tolerance [m]

    Returns
    -------
    ind_rng : int
        The range index

    """
    dist = np.abs(rng_vec-rng)
    ind_rng = np.argmin(dist)
    if dist[ind_rng] > rng_tol:
        return None

    return ind_rng


def get_target_elevations(radar_in):
    """
    Gets RHI target elevations

    Parameters
    ----------
    radar_in : Radar object
        current radar object

    Returns
    -------
    target_elevations : 1D-array
        Azimuth angles
    el_tol : float
        azimuth tolerance
    """
    sweep_start = radar_in.sweep_start_ray_index['data'][0]
    sweep_end = radar_in.sweep_end_ray_index['data'][0]
    target_elevations = np.sort(
        radar_in.elevation['data'][sweep_start:sweep_end+1])
    el_tol = np.median(target_elevations[1:]-target_elevations[:-1])

    return target_elevations, el_tol


def find_nearest_gate(radar, lat, lon, latlon_tol=0.0005):
    """
    Find the radar gate closest to a lat,lon point

    Parameters
    ----------
    radar : radar object
        the radar object
    lat, lon : float
        The position of the point
    latlon_tol : float
        The tolerance around this point

    Returns
    -------
    ind_ray, ind_rng : int
        The ray and range index
    azi, rng : float
        the range and azimuth position of the gate

    """
    # find gates close to lat lon point
    inds_ray_aux, inds_rng_aux = np.where(np.logical_and(
        np.logical_and(
            radar.gate_latitude['data'] < lat+latlon_tol,
            radar.gate_latitude['data'] > lat-latlon_tol),
        np.logical_and(
            radar.gate_longitude['data'] < lon+latlon_tol,
            radar.gate_longitude['data'] > lon-latlon_tol)))

    if inds_ray_aux.size == 0:
        warn('No data found at point lat '+str(lat)+' +- ' +
             str(latlon_tol)+' lon '+str(lon)+' +- ' +
             str(latlon_tol)+' deg')

        return None, None, None, None

    # find closest latitude
    ind_min = np.argmin(np.abs(
        radar.gate_latitude['data'][inds_ray_aux, inds_rng_aux]-lat))
    ind_ray = inds_ray_aux[ind_min]
    ind_rng = inds_rng_aux[ind_min]

    azi = radar.azimuth['data'][ind_ray]
    rng = radar.range['data'][ind_rng]

    return ind_ray, ind_rng, azi, rng


def find_neighbour_gates(radar, azi, rng, delta_azi=None, delta_rng=None):
    """
    Find the neighbouring gates within +-delta_azi and +-delta_rng

    Parameters
    ----------
    radar : radar object
        the radar object
    azi, rng : float
        The azimuth [deg] and range [m] of the central gate
    delta_azi, delta_rng : float
        The extend where to look for

    Returns
    -------
    inds_ray_aux, ind_rng_aux : int
        The indices (ray, rng) of the neighbouring gates

    """
    # find gates close to lat lon point
    if delta_azi is None:
        inds_ray = np.ma.arange(radar.azimuth['data'].size)
    else:
        azi_max = azi+delta_azi
        azi_min = azi-delta_azi
        if azi_max > 360.:
            azi_max -= 360.
        if azi_min < 0.:
            azi_min += 360.
        if azi_max > azi_min:
            inds_ray = np.where(np.logical_and(
                radar.azimuth['data'] < azi_max,
                radar.azimuth['data'] > azi_min))[0]
        else:
            inds_ray = np.where(np.logical_or(
                radar.azimuth['data'] > azi_min,
                radar.azimuth['data'] < azi_max))[0]
    if delta_rng is None:
        inds_rng = np.ma.arange(radar.range['data'].size)
    else:
        inds_rng = np.where(np.logical_and(
            radar.range['data'] < rng+delta_rng,
            radar.range['data'] > rng-delta_rng))[0]

    return inds_ray, inds_rng


def get_data_along_rng(radar, field_name, fix_elevations, fix_azimuths,
                       ang_tol=1., rmin=None, rmax=None):
    """
    Get data at particular (azimuths, elevations)

    Parameters
    ----------
    radar : radar object
        the radar object where the data is
    field_name : str
        name of the field to filter
    fix_elevations, fix_azimuths: list of floats
        List of elevations, azimuths couples [deg]
    ang_tol : float
        Tolerance between the nominal angle and the radar angle [deg]
    rmin, rmax: float
        Min and Max range of the obtained data [m]

    Returns
    -------
    xvals : 1D float array
        The ranges of each azi, ele pair
    yvals : 1D float array
        The values
    valid_azi, valid_ele : float arrays
        The azi, ele pairs

    """
    if rmin is None:
        rmin = 0.
    if rmax is None:
        rmax = np.max(radar.range['data'])

    rng_mask = np.logical_and(
        radar.range['data'] >= rmin, radar.range['data'] <= rmax)

    x = radar.range['data'][rng_mask]

    xvals = []
    yvals = []
    valid_azi = []
    valid_ele = []
    if radar.scan_type == 'ppi':
        for ele, azi in zip(fix_elevations, fix_azimuths):
            ind_sweep = find_ang_index(
                radar.fixed_angle['data'], ele, ang_tol=ang_tol)
            if ind_sweep is None:
                warn('No elevation angle found for fix_elevation '+str(ele))
                continue
            new_dataset = radar.extract_sweeps([ind_sweep])

            try:
                dataset_line = cross_section_ppi(
                    new_dataset, [azi], az_tol=ang_tol)
            except EnvironmentError:
                warn(' No data found at azimuth '+str(azi) +
                     ' and elevation '+str(ele))
                continue
            yvals.append(dataset_line.fields[field_name]['data'][0, rng_mask])
            xvals.append(x)
            valid_azi.append(dataset_line.azimuth['data'][0])
            valid_ele.append(dataset_line.elevation['data'][0])
    else:
        for ele, azi in zip(fix_elevations, fix_azimuths):
            ind_sweep = find_ang_index(
                radar.fixed_angle['data'], azi, ang_tol=ang_tol)
            if ind_sweep is None:
                warn('No azimuth angle found for fix_azimuth '+str(azi))
                continue
            new_dataset = radar.extract_sweeps([ind_sweep])

            try:
                dataset_line = cross_section_rhi(
                    new_dataset, [ele], el_tol=ang_tol)
            except EnvironmentError:
                warn(' No data found at azimuth '+str(azi) +
                     ' and elevation '+str(ele))
                continue
            yvals.extend(
                dataset_line.fields[field_name]['data'][0, rng_mask])
            xvals.extend(x)
            valid_azi.append(dataset_line.azimuth['data'][0])
            valid_ele.append(dataset_line.elevation['data'][0])

    return np.array(xvals), np.ma.array(yvals), valid_azi, valid_ele


def get_data_along_azi(radar, field_name, fix_ranges, fix_elevations,
                       rng_tol=50., ang_tol=1., azi_start=None,
                       azi_stop=None):
    """
    Get data at particular (ranges, elevations)

    Parameters
    ----------
    radar : radar object
        the radar object where the data is
    field_name : str
        name of the field to filter
    fix_ranges, fix_elevations: list of floats
        List of ranges [m], elevations [deg] couples
    rng_tol : float
        Tolerance between the nominal range and the radar range [m]
    ang_tol : float
        Tolerance between the nominal angle and the radar angle [deg]
    azi_start, azi_stop: float
        Start and stop azimuth angle of the data [deg]

    Returns
    -------
    xvals : 1D float array
        The ranges of each rng, ele pair
    yvals : 1D float array
        The values
    valid_rng, valid_ele : float arrays
        The rng, ele pairs

    """
    if azi_start is None:
        azi_start = np.min(radar.azimuth['data'])
    if azi_stop is None:
        azi_stop = np.max(radar.azimuth['data'])

    yvals = []
    xvals = []
    valid_rng = []
    valid_ele = []
    for rng, ele in zip(fix_ranges, fix_elevations):
        ind_rng = find_rng_index(radar.range['data'], rng, rng_tol=rng_tol)
        if ind_rng is None:
            warn('No range gate found for fix_range '+str(rng))
            continue

        if radar.scan_type == 'ppi':
            ind_sweep = find_ang_index(
                radar.fixed_angle['data'], ele, ang_tol=ang_tol)
            if ind_sweep is None:
                warn('No elevation angle found for fix_elevation ' +
                     str(ele))
                continue
            new_dataset = radar.extract_sweeps([ind_sweep])
        else:
            try:
                new_dataset = cross_section_rhi(radar, [ele], el_tol=ang_tol)
            except EnvironmentError:
                warn(
                    ' No data found at range '+str(rng) +
                    ' and elevation '+str(ele))
                continue
        if azi_start < azi_stop:
            azi_mask = np.logical_and(
                new_dataset.azimuth['data'] >= azi_start,
                new_dataset.azimuth['data'] <= azi_stop)
        else:
            azi_mask = np.logical_or(
                new_dataset.azimuth['data'] >= azi_start,
                new_dataset.azimuth['data'] <= azi_stop)
        yvals.extend(
            new_dataset.fields[field_name]['data'][azi_mask, ind_rng])
        xvals.extend(new_dataset.azimuth['data'][azi_mask])
        valid_rng.append(new_dataset.range['data'][ind_rng])
        valid_ele.append(new_dataset.elevation['data'][0])

    return np.array(xvals), np.ma.array(yvals), valid_rng, valid_ele


def get_data_along_ele(radar, field_name, fix_ranges, fix_azimuths,
                       rng_tol=50., ang_tol=1., ele_min=None,
                       ele_max=None):
    """
    Get data at particular (ranges, azimuths)

    Parameters
    ----------
    radar : radar object
        the radar object where the data is
    field_name : str
        name of the field to filter
    fix_ranges, fix_azimuths: list of floats
        List of ranges [m], azimuths [deg] couples
    rng_tol : float
        Tolerance between the nominal range and the radar range [m]
    ang_tol : float
        Tolerance between the nominal angle and the radar angle [deg]
    ele_min, ele_max: float
        Min and max elevation angle [deg]

    Returns
    -------
    xvals : 1D float array
        The ranges of each rng, ele pair
    yvals : 1D float array
        The values
    valid_rng, valid_ele : float arrays
        The rng, ele pairs

    """
    if ele_min is None:
        ele_min = np.min(radar.elevation['data'])
    if ele_max is None:
        ele_max = np.max(radar.elevation['data'])

    yvals = []
    xvals = []
    valid_rng = []
    valid_azi = []
    for rng, azi in zip(fix_ranges, fix_azimuths):
        ind_rng = find_rng_index(radar.range['data'], rng, rng_tol=rng_tol)
        if ind_rng is None:
            warn('No range gate found for fix_range '+str(rng))
            continue

        if radar.scan_type == 'ppi':
            try:
                new_dataset = cross_section_ppi(radar, [azi], az_tol=ang_tol)
            except EnvironmentError:
                warn(
                    ' No data found at range '+str(rng) +
                    ' and elevation '+str(azi))
                continue
        else:
            ind_sweep = find_ang_index(
                radar.fixed_angle['data'], azi, ang_tol=ang_tol)
            if ind_sweep is None:
                warn('No azimuth angle found for fix_azimuth '+str(azi))
                continue
            new_dataset = radar.extract_sweeps([ind_sweep])

        ele_mask = np.logical_and(
            new_dataset.elevation['data'] >= ele_min,
            new_dataset.elevation['data'] <= ele_max)
        yvals.extend(
            new_dataset.fields[field_name]['data'][ele_mask, ind_rng])
        xvals.extend(new_dataset.elevation['data'][ele_mask])
        valid_rng.append(new_dataset.range['data'][ind_rng])
        valid_azi.append(new_dataset.elevation['data'][0])

    return np.array(xvals), np.ma.array(yvals), valid_rng, valid_azi


def find_ang_index(ang_vec, ang, ang_tol=0.):
    """
    Find the angle index corresponding to a particular fixed angle

    Parameters
    ----------
    ang_vec : float array
        The angle data array where to look for
    ang : float
        The angle to search
    ang_tol : float
        Tolerance [deg]

    Returns
    -------
    ind_ang : int
        The angle index

    """
    dist = np.abs(ang_vec-ang)
    ind_ang = np.argmin(dist)
    if dist[ind_ang] > ang_tol:
        return None

    return ind_ang


def _create_qvp_object(radar, field_names, qvp_type='qvp', start_time=None,
                       hmax=10000., hres=200.):
    """
    Creates a QVP object containing fields from a radar object that can
    be used to plot and produce the quasi vertical profile

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field_names : list of strings
        Radar fields to use for QVP calculation.
    qvp_type : str
        Type of QVP. Can be qvp, rqvp, evp
    start_time : datetime object
        the QVP start time
    hmax : float
        The maximum height of the QVP [m]. Default 10000.
    hres : float
        The QVP range resolution [m]. Default 50

    Returns
    -------
    qvp : Radar-like object
        A quasi vertical profile object containing fields
        from a radar object

    """
    qvp = deepcopy(radar)

    # prepare space for field
    qvp.fields = dict()
    for field_name in field_names:
        qvp.add_field(field_name, deepcopy(radar.fields[field_name]))
        qvp.fields[field_name]['data'] = np.array([], dtype='float64')

    # fixed radar objects parameters
    qvp.range['data'] = np.arange(hmax/hres)*hres+hres/2.
    qvp.ngates = len(qvp.range['data'])

    if start_time is None:
        qvp.time['units'] = radar.time['units']
    else:
        qvp.time['units'] = make_time_unit_str(start_time)

    qvp.time['data'] = np.array([], dtype='float64')
    qvp.scan_type = qvp_type
    qvp.sweep_number['data'] = np.array([0], dtype='int32')
    qvp.nsweeps = 1
    qvp.sweep_mode['data'] = np.array([qvp_type])
    qvp.sweep_start_ray_index['data'] = np.array([0], dtype='int32')

    if qvp.rays_are_indexed is not None:
        qvp.rays_are_indexed['data'] = np.array(
            [qvp.rays_are_indexed['data'][0]])
    if qvp.ray_angle_res is not None:
        qvp.ray_angle_res['data'] = np.array([qvp.ray_angle_res['data'][0]])

    if qvp_type in ('rqvp', 'evp', 'vp'):
        qvp.fixed_angle['data'] = np.array([90.], dtype='float64')

    # ray dependent radar objects parameters
    qvp.sweep_end_ray_index['data'] = np.array([-1], dtype='int32')
    qvp.rays_per_sweep['data'] = np.array([0], dtype='int32')
    qvp.azimuth['data'] = np.array([], dtype='float64')
    qvp.elevation['data'] = np.array([], dtype='float64')
    qvp.nrays = 0

    return qvp


def _create_along_coord_object(radar, field_names, rng_values, fixed_angle,
                               mode, start_time=None):
    """
    Creates an along coord object containing fields from a radar object that
    can be used to plot and produce the time series along a coordinate

    Parameters
    ----------
    radar : Radar
        Radar object used.
    field_names : list of strings
        Radar fields to use for QVP calculation.
    rng_values : 1D-array
        The values to put in the range field
    fixed_angle : float
        the fixed angle
    mode : str
        The along coord mode, can be ALONG_AZI, ALONG_ELE, ALONG_RNG
    start_time : datetime object
        the acoord start time

    Returns
    -------
    acoord : Radar-like object
        An along coordinate object containing fields from a radar object

    """
    acoord = deepcopy(radar)

    # prepare space for field
    acoord.fields = dict()
    for field_name in field_names:
        acoord.add_field(field_name, deepcopy(radar.fields[field_name]))
        acoord.fields[field_name]['data'] = np.array([], dtype='float64')

    # fixed radar objects parameters
    acoord.range['data'] = rng_values
    acoord.ngates = len(acoord.range['data'])

    if start_time is None:
        acoord.time['units'] = radar.time['units']
    else:
        acoord.time['units'] = make_time_unit_str(start_time)

    acoord.time['data'] = np.array([], dtype='float64')
    acoord.scan_type = mode
    acoord.sweep_number['data'] = np.array([0], dtype='int32')
    acoord.nsweeps = 1
    acoord.sweep_mode['data'] = np.array([mode])
    acoord.sweep_start_ray_index['data'] = np.array([0], dtype='int32')

    if acoord.rays_are_indexed is not None:
        acoord.rays_are_indexed['data'] = np.array(
            [acoord.rays_are_indexed['data'][0]])
    if acoord.ray_angle_res is not None:
        acoord.ray_angle_res['data'] = np.array([acoord.ray_angle_res['data'][0]])

    acoord.fixed_angle['data'] = np.array([fixed_angle], dtype='float64')

    # ray dependent radar objects parameters
    acoord.sweep_end_ray_index['data'] = np.array([-1], dtype='int32')
    acoord.rays_per_sweep['data'] = np.array([0], dtype='int32')
    acoord.azimuth['data'] = np.array([], dtype='float64')
    acoord.elevation['data'] = np.array([], dtype='float64')
    acoord.nrays = 0

    return acoord


def _update_qvp_metadata(qvp, ref_time, lon, lat, elev=90.):
    """
    updates a QVP object metadata with data from the current radar volume

    Parameters
    ----------
    qvp : QVP object
        QVP object
    ref_time : datetime object
        the current radar volume reference time

    Returns
    -------
    qvp : QVP object
        The updated QVP object

    """
    start_time = num2date(0, qvp.time['units'], qvp.time['calendar'])
    qvp.time['data'] = np.append(
        qvp.time['data'], (ref_time - start_time).total_seconds())
    qvp.sweep_end_ray_index['data'][0] += 1
    qvp.rays_per_sweep['data'][0] += 1
    qvp.nrays += 1

    qvp.azimuth['data'] = np.ones((qvp.nrays, ), dtype='float64')*0.
    qvp.elevation['data'] = (
        np.ones((qvp.nrays, ), dtype='float64')*elev)

    qvp.gate_longitude['data'] = (
        np.ones((qvp.nrays, qvp.ngates), dtype='float64')*lon)
    qvp.gate_latitude['data'] = (
        np.ones((qvp.nrays, qvp.ngates), dtype='float64')*lat)
    qvp.gate_altitude['data'] = np.broadcast_to(
        qvp.range['data'], (qvp.nrays, qvp.ngates))

    return qvp


def _update_along_coord_metadata(acoord, ref_time, elevation, azimuth):
    """
    updates an along coordinate object metadata with data from the current
    radar volume

    Parameters
    ----------
    acoord : along coordinate object
        along coordinate object
    ref_time : datetime object
        the current radar volume reference time
    elevation, azimuth : 1D-array
        the elevation and azimuth value of the data selected

    Returns
    -------
    acoord : along coordinate object
        The updated along coordinate object

    """
    start_time = num2date(0, acoord.time['units'], acoord.time['calendar'])
    acoord.time['data'] = np.append(
        acoord.time['data'], (ref_time - start_time).total_seconds())
    acoord.sweep_end_ray_index['data'][0] += 1
    acoord.rays_per_sweep['data'][0] += 1
    acoord.nrays += 1

    acoord.azimuth['data'] = np.append(acoord.azimuth['data'], azimuth)
    acoord.elevation['data'] = np.append(acoord.elevation['data'], elevation)

    return acoord



"""
pyart.retrieve.visibility
==================

Functions for visibility and ground echoes estimation from a DEM.

"""

import numpy as np
import warnings

from itertools import product, repeat
from scipy.interpolate import interp1d
from functools import partial
from pyproj import Transformer, Proj
import multiprocessing as mp


from ..config import get_field_name, get_metadata, get_fillvalue
from ..core import antenna_vectors_to_cartesian as rad_to_cart

# These 2 are non picklable and need to defined as globals
nngrid = 0
transformer = 0

# Define worker function for parallelization on azimuth angles
def _worker_function(angles, ranges, dem_data, radar_params, 
                     numeric_params, atm_params):
    
    print(angles)
    rad_alt = radar_params['radar_altitude']
    freq = radar_params['freq']
    power = radar_params['power']
    gain = radar_params['gain']
    loss = radar_params['loss'] 
    beamwidth = radar_params['beamwidth']
    tau = radar_params['tau']
    
    atm_att = atm_params['atm_att']
    ke = atm_params['ke']
    mosotti = atm_params['mosotti']
    
    grid_res = numeric_params['grid_res']
    quad_pts_range = numeric_params['quad_pts_range']
    quad_pts_GH = numeric_params['GH_pts']
    GH_weights = numeric_params['GH_weights']
    sigma_0_method = numeric_params['sigma_0_method']
    
    # Get original length of the ray, before oversampling (see below)
    original_range_length = int(len(ranges) / quad_pts_range)
    visib_integ = np.ma.zeros((original_range_length,)) 
    rcs_integ = np.ma.zeros((original_range_length,)) 
    clutter_dBm_integ = np.ma.zeros((original_range_length,)) 
    clutter_dBZ_integ = np.ma.zeros((original_range_length,)) 
    
    i = 0
    for ph, pv in quad_pts_GH: # Loop on quadrature points
        # Get radar Cartesian coords at quadrature pt elevation and azimuth
        xr, yr, zr = rad_to_cart(ranges, angles[0] + ph, angles[1] + pv, 
                                 ke = ke)
        # Add radar altitude to z coordinates
        zr = zr[0] + rad_alt
        
        # Project to DEM coordinates and get corresponding grid indexes
        xr_proj, yr_proj = transformer.transform(xr, yr)
        i_idx,j_idx = nngrid.get_idx(xr_proj, yr_proj)
        
        
        # Create ray and compute relevant variabls
        ray = _Ray(angles[0], ranges, i_idx, j_idx, dem_data)
        ray.calc_dem_bent(dem_data, zr)
        ray.calc_slope_and_aspect(dem_data, grid_res)
        ray.calc_visibility()
        if ray.topography_encounter_idx != None:
            ray.calc_theta(onlytopo = True)
            ray.calc_incident_ang(onlytopo = True)
            ray.calc_eff_area(grid_res, onlytopo = True)
            ray.calc_sigma_0(sigma_0_method, freq, onlytopo = True)
            ray.calc_rcs()
            ray.calc_clutter_mW(power, gain, loss, freq, atm_att)
            ray.calc_clutter_Z(freq, mosotti, tau, beamwidth)
 
        # aggregate
        visib = np.nanmean(np.reshape(ray.visibility,
                                      (original_range_length, 
                                       quad_pts_range)), 
                                       axis=1)
        rcs = np.nanmean(np.reshape(ray.rcs,
                                      (original_range_length, 
                                       quad_pts_range)), 
                                       axis=1)
        clutter_W = np.nanmean(np.reshape(ray.clutter_W,
                                      (original_range_length, 
                                       quad_pts_range)), 
                                       axis=1)
        clutter_Z = np.nanmean(np.reshape(ray.clutter_Z,
                                      (original_range_length, 
                                       quad_pts_range)), 
                                       axis=1)
         
        # sum up with quadrature GH_weights
        visib_integ += visib * GH_weights[i]
        rcs_integ += rcs * GH_weights[i]
        clutter_dBm_integ += clutter_W * GH_weights[i]
        clutter_dBZ_integ += clutter_Z * GH_weights[i]
        i += 1
    
    clutter_dBm_integ = 10 * np.log10(clutter_dBm_integ) # Convert mW -> dBm
    clutter_dBZ_integ = 10 * np.log10(clutter_dBZ_integ) # Convert mm6m-3 -> dBZ
    return(visib_integ, rcs_integ, clutter_dBm_integ, clutter_dBZ_integ)

class _IndexNNGrid(object):
    """
    A class for fast interpolation in a 2D grid. It is able to find the closest
    point in a 2D grid for an arbitrary set of coordinates.

    Attributes
    ----------
    interp_i : Scipy 1D Interpolator
        Returns the i index (rows) of the closest point in the grid for a
        given x coordinate.
    interp_j : Scipy 1D Interpolator
        Returns the j index (rows) of the closest point in the grid for a
        given y coordinate.
    """
    
    def __init__(self, x, y):
        """Creates the interpolators from the Grid coordinates x and y"""
        I = np.arange(len(x))
        J = np.arange(len(y))
        self.interp_i = interp1d(x, I, bounds_error = False, kind = 'nearest') 
        self.interp_j = interp1d(y, J, bounds_error = False, kind = 'nearest') 
        
    def get_idx(self, xi, yi):
        """Returns the indexes of the closest grid points for a set of x 
        and y coordinates """
        i = self.interp_i(xi).astype(int)[0]
        j = self.interp_j(yi).astype(int)[0]
        i = np.ma.array(i, mask = i < 0)
        j = np.ma.array(j, mask = j < 0)
        return i, j    

class _Ray(object): 
    """
    A class that computes visibility variables on a single radar ray

    Attributes
    ----------
    i_idx : array
        Row indexes of all ray coordinates in the DEM grid.
    j_idx : array
        Column indexes of all ray coordinates in the DEM grid.
    valid_idx : bool array
        Boolean array that gives which ray indexes are valid (i.e. with
        a valid associated DEM value).
    i_idx_valid : array
        Contains only the valid row indexes.
    j_idx_valid : array
        Contains only the valid column indexes.
    dem : array
        Terrain altitude along ray
    dem_bent : array
        Terrain altitude corrected for radar beam height along ray
    slope : array
        Terrain slope along ray
    aspect : array
        Terrain aspect along ray
    theta : array
        Incidence elevation angle at terrain altitude along ray
    min_vis_theta : array
        Minimum elevation angle with sufficient visibility along ray
    min_vis_alt : array
        Minimum altitude with sufficient visibility along ray
    visib : array
        Radar visibility along ray ( 1 or 0)
    topography_encounter_idx : array
        Index along ray where the beam hits the topography
    """
    
    def __init__(self, az, ranges, i_idx, j_idx, dem_data):
        """
        Initializes the class
        
        Parameters
        ----------
        az : float
            Azimuth angle of that ray
        ranges : array
            Ranges in meters to all gates in the ray
        i_idx : array
            Row indexes of all ray coordinates in the DEM grid, as obtained
            with the _IndexNNGrid class.
        j_idx : array
            Column indexes of all ray coordinates in the DEM grid, as obtained
            with the _IndexNNGrid class.
        dem_data : array
            2D array that contains the DEM terrain altitude data (the grid
            for which to use i_idx, j_idx).
        """
        self.az = az
        self.ranges = ranges
        self.ranges_log = np.log10(ranges)
        
        self.i_idx = i_idx
        self.j_idx = j_idx
        self.calc_valid_gates(dem_data)
        
        self.i_idx_valid = self.i_idx[self.valid_idx]
        self.j_idx_valid = self.j_idx[self.valid_idx]
        self.N = len(self.valid_idx) # Nb of range gates
        
        self.dem_bent = np.zeros((self.N))
        self.slope = np.zeros((self.N))
        self.aspect = np.zeros((self.N))
        self.theta = np.zeros((self.N))
        self.min_vis_theta = np.zeros((self.N))
        self.min_vis_alt = np.zeros((self.N))
        self.area_eff = np.zeros((self.N))
        self.incident_ang = np.zeros((self.N))
        self.rcs = np.zeros((self.N))
        self.clutter_Z = np.zeros((self.N))
        self.clutter_W = np.zeros((self.N))
        self.sigma_0 = np.zeros((self.N))
        
    def calc_valid_gates(self, dem_data):
        """
        Calculates the valid gates of the ray 
        
        Parameters
        ----------
        dem_data : array
            2D array that contains the DEM terrain altitude data (the grid
            for which to use i_idx, j_idx)
        """
        # valid gates are those were i and j indexes are defined and DEM
        # data is available (no hole in DEM grid)
        valid = np.logical_and(~self.i_idx.mask, ~self.j_idx.mask)
        valid[self.i_idx == dem_data.shape[0] - 1] = False
        valid[self.j_idx == dem_data.shape[1] - 1] = False
        valid[self.i_idx == 0] = False
        valid[self.j_idx == 0] = False
        valid[valid] = ~dem_data[self.i_idx[valid], self.j_idx[valid]].mask
        
        self.valid_idx = valid
        
    def calc_dem_bent(self, dem_data, alt):
        """
        Bends the DEM data along the ray (i.e. subtracts the altitude of
        the radar gates from the terrain altitude).
        
        Parameters
        ----------        
        alt : array
            Altitude of all radar gates along the ray
        """
        self.dem = np.ma.zeros((self.N))
        self.dem[self.valid_idx] = (dem_data[self.i_idx_valid, 
                                             self.j_idx_valid])
        self.dem.mask = ~self.valid_idx
        self.dem_bent = self.dem - alt

        
    def calc_slope_and_aspect(self, dem_data, grid_res, onlytopo = False):
        """
        Calculates the slope and aspect from the DEM data using Sobel filters
        to compute the gradient in X and Y directions.
        
        Parameters
        ----------
        dem_data : array
            2D array that contains the DEM terrain altitude data (the grid
            for which to use i_idx, j_idx)
        
        grid_res : int or float
            Spatial resolution of the grid on which the DEM is defined, for
            example 25 meters.
        
        idx_along_ray : array, optional
            Array of 1D indexes along ray where to compute the slope and aspect
            if not provided it will be computed all along the ray
        """
        # simplify notation
        i = self.i_idx
        j = self.j_idx
        valid = self.valid_idx
        if onlytopo:
            i = i[self.topography_encounter_idx]
            j = j[self.topography_encounter_idx]
            valid = valid[self.topography_encounter_idx]
            
        d = dem_data
        
        # Sobel filter in Y direction
        gy = np.ma.zeros((self.N)) + np.nan
        gy[self.valid_idx] = (d[i[valid]-1,j[valid]-1] + 
                   2 * d[i[valid],j[valid]-1] + 
                   d[i[valid]+1,j[valid]-1] -
                   d[i[valid]-1,j[valid]+1] -
                   2 * d[i[valid],j[valid]+1] -
                   d[i[valid]+1,j[valid]+1])/(8 * grid_res)
        
        # Sobel filter in X direction
        gx = np.ma.zeros((self.N)) + np.nan
        gx[valid] = (d[i[valid]+1,j[valid]-1] + 
                   2 * d[i[valid]+1,j[valid]] + 
                   d[i[valid]+1,j[valid]+1] -
                   d[i[valid]-1,j[valid]-1] -
                   2 * d[i[valid]-1,j[valid]] -
                   d[i[valid]-1,j[valid]+1])/(8 * grid_res)
        
        slope = np.ma.zeros((self.N)) 
        slope[valid] = np.arctan(np.sqrt(gy[valid]**2 + 
                                         gx[valid]**2)) * 180 / np.pi
        aspect = np.ma.zeros((self.N)) + np.nan
        aspect[valid] = (np.arctan2(gy[valid],
                                    -gx[valid]) + np.pi) * 180 / np.pi
        
        slope.mask = ~valid
        aspect.mask = ~valid
        self.slope = slope
        self.aspect = aspect
        
    def calc_theta(self, onlytopo = False):
        """
        Calculates the theta angle, i.e. the elevation angle at which the radar
        ray touches the topography
        
        Parameters
        ----------
        ranges : array
            range of every gate from the radar.
        
        """
    
        if not hasattr(self, 'dem_bent'):
            warnings.warn('Please use bend_dem() function first...')
            return None

        i = self.i_idx
        j = self.j_idx
        valid = self.valid_idx
        if onlytopo:
            i = np.array([i[self.topography_encounter_idx]])
            j = np.array([j[self.topography_encounter_idx]])
            valid = np.array([self.topography_encounter_idx])
            
        theta = np.ma.zeros((self.N))
        theta[valid] = (np.arctan2(self.dem_bent[valid], self.ranges[valid]) *
                                180 / np.pi)
        self.theta = theta
    
    def calc_min_theta(self, beamwidth):
        """
        Calculates the minimum theta angle, at which there is sufficient 
        visibility. 
        """
        if not hasattr(self, 'theta'):
           warnings.warn('Please use calc_theta() function first...')
           return None
       
        min_vis_theta = np.zeros((self.N)) + np.nan 
        current = self.theta[0]
        for i in range(1, self.N):
            if np.isnan(self.theta[i]):
                min_vis_theta[i] = np.nan
            elif self.theta[i] > current:
                current = self.theta[i]
        min_vis_theta[i] = current + beamwidth / 2.
        
        self.min_vis_theta = min_vis_theta

        
    def calc_min_vis_alt(self, ke):
        """
        Calculates the minimum visible altitude along the ray
        
        Parameters
        ----------
        ranges : array
            range of every gate from the radar.
         beamwidth : float, optional
            3dB beamwidth of the antenna pattern.
        ke : float, optional
            Effective radius scale factor 
        
        """

        if not hasattr(self, 'min_vis_theta'):
            warnings.warn('Please use calc_min_theta() function first...')
            return None
        
        # elevation angle in radians.
        theta_e = (self.min_vis_theta) * np.pi / 180.0   
        R = 6371.0 * 1000.0 * ke     # effective radius of earth in meters.
    
        self.min_vis_alt = (self.ranges ** 2 + R ** 2 + 2.0 * self.ranges * R *
                                    np.sin(theta_e)) ** 0.5 - R

    def calc_visibility(self):
        """
        Calculates the visibility along the ray : all gates after first 
        encounter with topography have zero visibility, as well as the index
        along the ray where the beam first "hits" the topography
        """
        if not hasattr(self, 'dem_bent'):
            warnings.warn('Please use bend_dem() function first...')
            return None
        
        visib = np.ma.ones((self.N))
        idx_pos = np.where(self.dem_bent > 0)[0]
        if  len(idx_pos):
            visib[idx_pos[0]:] = 0
            self.topography_encounter_idx = idx_pos[0]
        else:
            self.topography_encounter_idx = None
            
        visib.mask = self.dem_bent.mask

        self.visibility = visib

        
    def calc_incident_ang(self, onlytopo = False):
        if not hasattr(self, 'theta'):
            warnings.warn('Please use calc_theta() function first...')
            return None
        if not hasattr(self, 'slope'):
            warnings.warn('Please use calc_slope_aspect() function first...')
            return None

        i = self.i_idx
        j = self.j_idx
        valid = self.valid_idx
        if onlytopo:
            i = np.array([i[self.topography_encounter_idx]])
            j = np.array([j[self.topography_encounter_idx]])
            valid = np.array([self.topography_encounter_idx])
            
        slope  = self.slope[valid] * np.pi / 180.0
        aspect = self.aspect[valid] * np.pi / 180.0
        zenith = (90. - self.theta[valid]) * np.pi / 180.0 
        az     = self.az * np.pi / 180.0 
        
        inc_ang = np.zeros((self.N))
        inc_ang[valid] = np.arccos(-( np.sin(slope) * np.sin(zenith) * 
             (np.sin(aspect) * np.sin(az) + np.cos(aspect) * np.cos(az)) + 
                     np.cos(slope) * np.cos(zenith))) * 180 / np.pi
        self.incident_ang = inc_ang

        
    def calc_eff_area(self, grid_res, onlytopo = False):
        if not hasattr(self, 'slope'):
            warnings.warn('Please use calc_slope_aspect() function first...')
            return None
        i = self.i_idx
        j = self.j_idx
        valid = self.valid_idx
        if onlytopo:
            i = np.array([i[self.topography_encounter_idx]])
            j = np.array([j[self.topography_encounter_idx]])
            valid = np.array([self.topography_encounter_idx])
            
        area_eff = np.zeros((self.N))
        area_eff[valid] = grid_res**2 / np.cos(self.slope[valid] *
                                                np.pi / 180.0)
        self.area_eff = area_eff
        
    def calc_rcs(self):
        if not hasattr(self, 'area_eff'):
            warnings.warn('Please use calc_area_eff() function first...')
            return None
        if not hasattr(self, 'sigma_0'):
            warnings.warn('Please use calc_area_eff() function first...')
            return None
        
        self.rcs = self.area_eff * self.sigma_0
        
    def calc_clutter_mW(self, freq, Pt, G, L, atm_att):
        if not hasattr(self, 'rcs'):
            warnings.warn('Please use calc_rcs() function first...')
        lambd = 3. / (freq * 10.) 
        pconst = (10 * np.log10(Pt) + 2 * G + 20 * np.log10(lambd) - L  - 
                  30 * np.log10(4 * np.pi))
       
        self.clutter_mW = 10**(0.1 * (pconst + 40 * self.ranges_log - 
                            2 * atm_att * self.ranges / 1000. + 
                            10 * np.log10(self.rcs)))

        
    def calc_clutter_Z(self, freq, mosotti, tau, beamwidth):
        if not hasattr(self, 'rcs'):
            warnings.warn('Please use calc_rcs() function first...')
        lambd = 3. / (freq * 10.) 
        dbzconst = (10 * np.log10(2) + 40 * np.log10(lambd) -
                   10 * np.log10(tau * 3e8) - 20 *np.log10(beamwidth) - 
                   60 * np.log10(np.pi) - 20 * np.log10(mosotti))
       
        convert_dbzm_to_dbz = 180. # 10*log10(1 m^6 / 1 mm^6) = 180

        self.clutter_Z = (self.rcs - 20 * self.ranges_log + dbzconst + 
                          convert_dbzm_to_dbz)

    def calc_sigma_0(self, method = 'gabella', freq_ghz = 5.6,
                     onlytopo = False):
        """
        Calculates sigma_0: the ratio between the backscattering cross section 
        and the physical backscattering area
    
        Parameters
        ----------
        method : str, optional
            Method to use in the calculation of sigma_0, either "gabella" or
            "delrieu"
        freq_ghz : float, optional
            Frequency of the radar in GHz, is used only for the Delrieu method
      
        """
        if not hasattr(self, 'incident_ang'):
            warnings.warn('Please use calc_incident_ang() function first...')
            return None
        
        i = self.i_idx
        j = self.j_idx
        valid = self.valid_idx
        if onlytopo:
            i = np.array([i[self.topography_encounter_idx]])
            j = np.array([j[self.topography_encounter_idx]])
            valid = np.array([self.topography_encounter_idx])
            
        sigma_0 = np.zeros((self.N))[valid]
        inc_ang = self.incident_ang[valid]
        inc_angr = inc_ang * np.pi / 180.0
        if method == 'gabella':
            # GABELLA/PERONA 1997 (GECS)
            te1 = 80.     # Maximum Angle of "Plateau" Region      [deg]
            te2 = 87.     # Dimming Start Angle at Grazing Angles  [deg]
            te3 = 88.     # Border Incidence Angle                 [deg]
            k1 = 0.1      # Offset Index
            k2 = 1.       # Dimming Index near Grazing Angles
            bbdeg = 2.75  # Tuning Angle in Exp Rising above Border Angle [deg]

            te1r = te1 * np.pi / 180.0
            te2r = te2 * np.pi / 180.0
            te3r = te3 * np.pi / 180.0
            bbdegr = bbdeg * np.pi / 180.0

            factor =  k1 * np.cos(te2r) * ((np.pi / 2. - te2r) / 
                                           (np.pi / 2.-te1r) ) ** k2
            
            ind0 = np.where(inc_ang <= te1)[0]
            if len(ind0):
                sigma_0[ind0] = k1 * np.cos(inc_angr[ind0])
            
            ind1 = np.where(np.logical_and(inc_ang > te1, inc_ang <= te2))[0]
            if len(ind1):
                iang = inc_angr[ind1]
                sigma_0[ind1] =  k1 * np.cos(iang) * ((np.pi / 2.
                                                              - iang) / 
                                               (np.pi / 2.-te1r) ) ** k2

            ind2 = np.where(np.logical_and(inc_ang > te2, inc_ang <= te3))[0]
            if len(ind2):
                sigma_0[ind2] = factor

            ind3 = np.where(np.logical_and(inc_ang > te3, inc_ang <= 90))[0]
            if len(ind3):
                iang = inc_angr[ind3]
                sigma_0[ind3] = factor * np.exp((iang - te3r) / bbdegr)

            ind4 = np.where(inc_ang > 90)
            if len(ind4):
                sigma_0[ind4] = factor * np.exp((np.pi / 2. - te3r) 
                                                       / bbdegr)
            
        elif method == 'delrieu':
            # DELRIEU 1995 (X-BAND) / SKOLNIK 1990
            lim_ang_del = 63.75  # Borderline Angle separating curve branche
            # First Branch  [0..lim_ang_del]
            a1 = -9.1    # Skolnik's Model "a1" parameter [dB]
            b1 = -0.12   # Skolnik's Model "b1" parameter [dB/deg]
            c1 = 0.25    # Skolnik's Model "c1" parameter [dB/GHz]
            d1 = 0.      # Skolnik's Model "d1" parameter [dB/(deg*GHz))]
            # Second Branch  [lim_ang_del..90]
            a1_2 = 12.93 # Skolnik's Model "a1" parameter [dB]
            b1_2 = -0.37 # Skolnik's Model "b1" parameter [dB/deg]
            c1_2 = 0.    # Skolnik's Model "c1" parameter [dB/GHz]
            d1_2 = 0.    # Skolnik's Model "d1" parameter [dB/(deg*GHz))]

            sigma_db = np.zeros((self.N))[valid]

            ind = np.where( inc_ang < lim_ang_del)[0]
            if len(ind):
                iang = inc_angr[ind]
                sigma_db[ind] = (a1 + b1 * iang + c1 * freq_ghz +
                                 d1 * freq_ghz * iang)
            ind = np.where( inc_ang >= lim_ang_del)[0]
            if len(ind):
               iang = inc_angr[ind]
               sigma_db[ind] = (a1_2 + b1_2 * iang + c1_2 * freq_ghz +
                                 d1_2 * freq_ghz * iang) 

            sigma_0 = 10 ** (sigma_db / 10.)
        else:
            warnings.warn('Invalid method for sigma_0: use "gabella" or '+\
                          '"delrieu"')
            return
        
        self.sigma_0[valid] = sigma_0
    
    
        
def _GH_quadrature(npts_el, npts_az, beamwidth):
    """
    Retrieves the Gauss-Hermite quadrature points and GH_weights for Gaussian
    antenna integration in azimuthal and elevational directions.

    Parameters
    ----------
    npts_el : int
        Number of quadrature points in the elevational direction.
    npts_az : int
        Number of quadrature points in the azimuthal direction.        
    beamwidth : float, optional
        3dB beamwidth of the antenna pattern.

        
    Returns
    -------
    quad_pts : array
        Quadrature points as an N x 2 array that contains the azimuthal
        offset in the first column and the elevational column in the second.
        The number of rows is equal to npts_el x npts_az.
    GH_weights : array
        Normalized quadrature GH_weights of all quadrature points.

    References
    ----------
    Wolfensberger, D. and Berne, A.: From model to radar variables: a new
    forward polarimetric radar operator for COSMO, Atmos. Meas. Tech., 11,
    3883–3916, https://doi.org/10.5194/amt-11-3883-2018, 2018. 
    """
    
    sigma = beamwidth/(2 * np.sqrt(2 * np.log(2)))
    
    pts_az, weights_az = np.polynomial.hermite.hermgauss(npts_az)
    pts_az *= sigma
    
    pts_el, weights_el = np.polynomial.hermite.hermgauss(npts_el)
    pts_el *= sigma
    
    # Create 2D array of GH_weights
    GH_weights = np.outer(weights_az * sigma, weights_el * sigma)
    GH_weights *= np.abs(np.cos(np.deg2rad(pts_el)))
    
    sum_weights = np.sum(GH_weights.ravel())
    GH_weights /= sum_weights # Normalize GH_weights
    
    quad_pts = list(product(pts_az,pts_el))
    GH_weights = GH_weights.ravel()
    
    return quad_pts, GH_weights

    
def visibility_processing(radar, dem_grid,
                          frequency, 
                          beamwidth, 
                          tau,
                          power,
                          gain,
                          loss,
                          fill_value = None, 
                          terrain_altitude_field = None,
                          bent_terrain_altitude_field = None,
                          terrain_slope_field = None,
                          terrain_aspect_field = None,
                          theta_angle_field = None,
                          visibility_field = None,
                          min_vis_altitude_field = None,
                          min_vis_theta_field = None,
                          incident_angle_field = None,
                          sigma_0_field = None,
                          rcs_ground_clutter_field = None,
                          dBm_ground_clutter_field = None,
                          dBZ_ground_clutter_field = None,
                          atm_att = 0.05,
                          mosotti = 0.9644,
                          ke = 4/3., 
                          sigma_0_method = 'gabella',
                          quad_pts_range = 1,
                          quad_pts_az = 9, 
                          quad_pts_el = 9, 
                          parallel = True):
            
    # parse fill value
    if fill_value is None:
        fill_value = get_fillvalue()

    # parse field names
    if terrain_altitude_field is None:
        terrain_altitude_field = get_field_name('terrain_altitude')
    if bent_terrain_altitude_field is None:
        bent_terrain_altitude_field = get_field_name('bent_terrain_altitude')
    if terrain_slope_field is None:
        terrain_slope_field = get_field_name('terrain_slope')
    if terrain_aspect_field is None:
        terrain_aspect_field = get_field_name('terrain_aspect')
    if theta_angle_field is None:
        theta_angle_field = get_field_name('theta_angle')
    if visibility_field is None:
        visibility_field = get_field_name('visibility')
    if min_vis_altitude_field is None:
        min_vis_altitude_field = get_field_name('min_vis_altitude')
    if incident_angle_field is None:
        incident_angle_field = get_field_name('incident_angle')    
    if sigma_0_field is None:
        sigma_0_field = get_field_name('sigma_0')   
    if rcs_ground_clutter_field is None:
        rcs_ground_clutter_field = get_field_name('rcs_ground_clutter')   
    if dBm_ground_clutter_field is None:
        dBm_ground_clutter_field = get_field_name('dBm_ground_clutter')   
    if dBZ_ground_clutter_field is None:
        dBZ_ground_clutter_field = get_field_name('dBZ_ground_clutter')   
        
    # Define aeqd projection for radar local Cartesian coords
    pargs = Proj(proj="aeqd", lat_0 = radar.latitude['data'][0], 
                 lon_0 = radar.longitude['data'][0], datum = "WGS84", 
                 units="m")    
    
    # Define coordinate transform: (local radar Cart coords) -> (DEM coords)
    global transformer
    transformer = Transformer.from_proj(pargs, dem_grid.projection)
    
    # Get quadrature pts and GH_weights in az,el directions
    GH_pts, GH_weights = _GH_quadrature(quad_pts_el, quad_pts_az, beamwidth)
    
    # Create grid interpolator
    global nngrid
    nngrid = _IndexNNGrid(dem_grid.x['data'], dem_grid.y['data'])
    
    # Initialize output fields
    
    nazimuth = len(radar.azimuth['data'])
    ngates = len(radar.range['data'])
    
    visibility = np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    theta =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    dem_bent = np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    terrain_slope =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    terrain_aspect =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    min_vis_theta =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    min_vis_altitude =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    incident_angle =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    sigma_0 =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    clutter_rcs =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    clutter_dBZ =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    clutter_dBm =  np.ma.zeros([nazimuth, ngates], fill_value = fill_value)
    ##
    
    
    dem_data = dem_grid.fields[terrain_altitude_field]['data']
    # In the first step we estimate the variables at topography level but
    # only over the radar coordinates        
    for i, az in enumerate(radar.get_azimuth(0)): # Loop on az angles
    
        # Get radar Cartesian coords at elevation 0
        xr, yr, zr = rad_to_cart(radar.range['data'],
                                 az, 0,
                                 ke = ke)
        # Add radar altitude to z coordinates
        zr = zr[0] + radar.altitude['data']
        
        # Project to DEM coordinates and get corresponding grid indexes
        xr_proj, yr_proj = transformer.transform(xr,yr)
        i_idx,j_idx = nngrid.get_idx(xr_proj,yr_proj)
        
        # Create ray and compute relevant variabls
        ray = _Ray(az, radar.range['data'], i_idx, j_idx, dem_data)
        ray.calc_dem_bent(dem_data, zr)
        ray.calc_slope_and_aspect(dem_data, dem_grid.metadata['resolution'])
        ray.calc_theta()
        ray.calc_min_theta(beamwidth)
        ray.calc_min_vis_alt( ke)
        ray.calc_incident_ang()
        ray.calc_sigma_0(freq_ghz = frequency)
        
        theta[i] = ray.theta
        min_vis_theta[i] = ray.min_vis_theta
        min_vis_altitude[i] = ray.min_vis_alt
        dem_bent[i] = ray.dem_bent
        terrain_slope[i] = ray.slope
        terrain_aspect[i] = ray.aspect
        sigma_0[i] = ray.sigma_0
        incident_angle[i] = ray.incident_ang
        
    theta_dic = get_metadata(theta_angle_field)
    theta_dic['data'] = theta
    
    slope_dic = get_metadata(terrain_slope_field)
    slope_dic['data'] = terrain_slope
    
    aspect_dic = get_metadata(terrain_aspect_field)
    aspect_dic['data'] = terrain_aspect
    
    min_vis_theta_dic = get_metadata(min_vis_theta_field)
    min_vis_theta_dic['data'] = min_vis_theta

    min_vis_altitude_dic = get_metadata(min_vis_altitude_field)
    min_vis_altitude_dic['data'] = min_vis_altitude  
    
    bent_terrain_altitude_dic = get_metadata(bent_terrain_altitude_field)
    bent_terrain_altitude_dic['data'] = dem_bent  

    incident_angle_dic = get_metadata(incident_angle_field)
    incident_angle_dic['data'] = incident_angle  
    
    sigma_0_dic = get_metadata(sigma_0_field)
    sigma_0_dic['data'] = sigma_0  
    
    if quad_pts_range >= 3:
        """ 
        Interpolate range array based on how many quadrature points in range
        are wanted (at least 3)
        For example if quad_pts_range = 5 and original range array is 
        [25,75,125] (bin centers), it will give [0,25,50,50,75,100,125,150]
        i.e. bin 0 - 50 (with center 25) is decomposed into ranges [0,25,50]
        """
        ranges = radar.range['data']
        nrange = len(radar.range['data'])
        dr = np.mean(np.diff(radar.range['data'])) # range res
        intervals = np.arange(ranges[0] - dr / 2, dr * nrange + dr / 2, dr)
        range_resampled = []
        for i in range(len(intervals) - 1):
            range_resampled.extend(np.linspace(intervals[i], intervals[i + 1],
                                               quad_pts_range))
    else:
        # No resampling
        range_resampled = radar.range['data']
    print('done')
    
    # create parallel computing instance
    if parallel:
        pool = mp.Pool(processes=mp.cpu_count(), maxtasksperchild=1)
        map_ = pool.map
    else:
        map_ = map
        
    # Create parameter dictionaries for workers
    radar_params = {}
    radar_params['radar_altitude'] = radar.altitude['data']
    radar_params['freq'] = frequency
    radar_params['power'] = power
    radar_params['gain'] = gain
    radar_params['loss'] = loss
    radar_params['beamwidth'] = beamwidth
    radar_params['tau'] = tau
    
    atm_params = {}
    atm_params['atm_att'] = atm_att
    atm_params['ke'] = ke
    atm_params['mosotti'] = mosotti
    
    numeric_params = {}
    numeric_params['grid_res'] = dem_grid.metadata['resolution']
    numeric_params['quad_pts_range'] = quad_pts_range
    numeric_params['GH_pts'] = GH_pts
    numeric_params['GH_weights'] = GH_weights
    numeric_params['sigma_0_method'] = sigma_0_method
    
    # Loop on fixed angles : el for ppi, az for rhi
    idx = 0
    for i, fixangle in enumerate(radar.fixed_angle['data']): 
        # Create partial worker func that takes only angle as input
        if radar.scan_type == 'ppi':
            angles = list((zip(radar.get_azimuth(i), repeat(fixangle))))
        elif radar.scan_type == 'rhi':
            angles = list((zip(repeat(fixangle), radar.get_elevation(i))))
            
        partialworker = partial(_worker_function, 
                                ranges = range_resampled,
                                dem_data = dem_data,
                                radar_params = radar_params,
                                atm_params = atm_params,
                                numeric_params = numeric_params)
   
        results = list(map_(partialworker, angles))
     
        visibility[idx : idx + len(results), :] = np.array([r[0] for r 
                                                            in results])
        clutter_rcs[idx : idx + len(results), :] = np.array([r[1] for r 
                                                            in results])
        clutter_dBm[idx : idx + len(results), :] = np.array([r[2] for r
                                                            in results])
        clutter_dBZ[idx : idx + len(results), :] = np.array([r[3] for r 
                                                            in results])
        
        idx += len(results)
        
        if parallel:
            pool.close()
       
    visibility_dic = get_metadata(visibility_field)
    visibility_dic['data'] = visibility  

    rcs_ground_clutter_dic = get_metadata(rcs_ground_clutter_field)
    rcs_ground_clutter_dic['data'] = clutter_rcs  
    
    dBm_ground_clutter_dic = get_metadata(dBm_ground_clutter_field)
    dBm_ground_clutter_dic['data'] = clutter_dBm 
    
    dBZ_ground_clutter_dic = get_metadata(dBZ_ground_clutter_field)
    dBZ_ground_clutter_dic['data'] = clutter_dBZ  
    
    return (bent_terrain_altitude_dic, slope_dic, aspect_dic, 
            theta_dic, min_vis_theta_dic, min_vis_altitude_dic,
            visibility_dic, incident_angle_dic, sigma_0_dic,
            rcs_ground_clutter_dic, dBm_ground_clutter_dic,
            dBZ_ground_clutter_dic)



"""
pyart.io.write_grid_geotiff
===========================

Write a Py-ART Grid object to a GeoTIFF file.

.. autosummary::
    :toctree: generated/

    write_grid_geotiff
    _get_rgb_values
    _create_sld

"""

import os
import shutil

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from ..exceptions import MissingOptionalDependency
try:
    from osgeo import gdal
    IMPORT_FLAG = True
except ImportError:
    IMPORT_FLAG = False


def write_grid_geotiff(grid, filename, field, rgb=False, level=None,
                       cmap='viridis', vmin=0, vmax=75, color_levels=None,
                       warp=False, sld=False, use_doublequotes=False):
    """
    Write a Py-ART Grid object to a GeoTIFF file.

    The GeoTIFF can be the standard Azimuthal Equidistant projection used
    in Py-ART, or a lat/lon projection on a WGS84 sphere. The latter is
    typically more usable in web mapping applications. The GeoTIFF can
    contain a single float-point raster band, or three RGB byte raster bands.
    The former will require an SLD file for colorful display using standard
    GIS or web mapping software, while the latter will show colors
    "out-of-the-box" but lack actual data values. The function also can
    output an SLD file based on the user-specified inputs. User can specify
    the 2D vertical level to be output. If this is not specified, a 2D
    composite is created. User also can specify the field to output.

    This function requires GDAL Python libraries to be installed. These are
    available via conda; e.g., 'conda install gdal'

    Parameters
    ----------
    grid : pyart.core.Grid object
        Grid object to write to file.
    filename : str
        Filename for the GeoTIFF.
    field : str
        Field name to output to file.

    Other Parameters
    ----------------
    rbg : bool, optional
        True - Output 3-band RGB GeoTIFF

        False - Output single-channel, float-valued GeoTIFF. For display,
                likely will need an SLD file to provide a color table.

    level : int or None, optional
        Index for z-axis plane to output. None gives composite values
        (i.e., max in each vertical column).
    cmap : str or matplotlib.colors.Colormap object, optional
        Colormap to use for RGB output or SLD file.
    vmin : int or float, optional
        Minimum value to color for RGB output or SLD file.
    vmax : int or float, optional
        Maximum value to color for RGB output or SLD file.
    color_levels : int or None, optional
        Number of color levels in cmap. Useful for categorical colormaps
        with steps << 255 (e.g., hydrometeor ID).
    warp : bool, optional
        True - Use gdalwarp (called from command line using os.system)
               to warp to a lat/lon WGS84 grid.

        False - No warping will be performed. Output will be Az. Equidistant.

    sld : bool, optional
        True - Create a Style Layer Descriptor file (SLD) mapped to vmin/vmax
               and cmap. File is named same as output TIFF, except for .sld
               extension.

        False - Don't do this.

    use_doublequotes : bool, optional
        True - Use double quotes in the gdalwarp call (requires warp=True),
               which may help if that command is producing and error like:
               'Translating source or target SRS failed'.

        False - Use single quotes instead.

    """
    if not IMPORT_FLAG:
        raise MissingOptionalDependency(
            'GDAL not detected, GeoTIFF output failure!')

    if field not in grid.fields.keys():
        raise KeyError('Failed -', field, 'field not found in Grid object.')

    # Determine whether filename template already contains a suffix
    # If not, append an appropriate one.
    if '.' not in filename:
        name = filename
        end = 'tif'
        ofile = name + "." + end
    else:
        ofile = filename
    nz, ny, nx = grid.fields[field]['data'].shape
    dist = max(grid.x['data'])
    rangestep = grid.x['data'][1] - grid.x['data'][2]
    lat = grid.origin_latitude['data'][0]
    lon = grid.origin_longitude['data'][0]
    # Check if masked array; if so, fill missing data
    filled = np.ma.filled(grid.fields[field]['data'], fill_value=-32768)
    if level is None:
        data = np.amax(filled, 0)
    else:
        data = filled[level]
    data = data.astype(float)
    data[data == -32768] = np.nan

    iproj = 'PROJCS["unnamed",GEOGCS["WGS 84",DATUM["unknown",' + \
        'SPHEROID["WGS84",6378137,298.257223563]],' + \
        'PRIMEM["Greenwich",0],' + \
        'UNIT["degree",0.0174532925199433]],' + \
        'PROJECTION["Azimuthal_Equidistant"],' + \
        'PARAMETER["latitude_of_center",' + str(lat) + '],' + \
        'PARAMETER["longitude_of_center",' + str(lon) + '],' + \
        'PARAMETER["false_easting",0],' + \
        'PARAMETER["false_northing",0],' + \
        'UNIT["metre",1,AUTHORITY["EPSG","9001"]]]'
    out_driver = gdal.GetDriverByName("GTiff")

    # Output dataset depends on rgb flag
    if not rgb:
        # Single-channel, floating-point output
        dst_options = ['COMPRESS=LZW', 'ALPHA=YES']
        dst_ds = out_driver.Create(
            ofile, data.shape[1], data.shape[0], 1,
            gdal.GDT_Float32, dst_options)
    else:
        # Assign data RGB levels based on value relative to vmax/vmin
        rarr, garr, barr = _get_rgb_values(
            data, vmin, vmax, color_levels, cmap)
        dst_ds = out_driver.Create(ofile, data.shape[1],
                                   data.shape[0], 3, gdal.GDT_Byte)

    # Common Projection and GeoTransform
    dst_ds.SetGeoTransform([-dist, -rangestep, 0, dist, 0, rangestep])
    dst_ds.SetProjection(iproj)

    # Final output depends on rgb flag
    if not rgb:
        dst_ds.GetRasterBand(1).WriteArray(data[::-1, :])
    else:
        dst_ds.GetRasterBand(1).WriteArray(rarr[::-1, :])
        dst_ds.GetRasterBand(2).WriteArray(garr[::-1, :])
        dst_ds.GetRasterBand(3).WriteArray(barr[::-1, :])
    dst_ds.FlushCache()
    dst_ds = None

    if sld:
        _create_sld(cmap, vmin, vmax, ofile, color_levels)

    if warp:
        # Warps TIFF to lat/lon WGS84 projection that is more useful
        # for web mapping applications. Likely changes array shape.
        if use_doublequotes:
            os.system('gdalwarp -q -t_srs \"+proj=longlat +ellps=WGS84 ' +
                      '+datum=WGS84 +no_defs\" ' + ofile + ' ' +
                      ofile + '_tmp.tif')
        else:
            os.system('gdalwarp -q -t_srs \'+proj=longlat +ellps=WGS84 ' +
                      '+datum=WGS84 +no_defs\' ' + ofile + ' ' +
                      ofile + '_tmp.tif')
        shutil.move(ofile+'_tmp.tif', ofile)


def _get_rgb_values(data, vmin, vmax, color_levels, cmap):
    """
    Get RGB values for later output to GeoTIFF, given a 2D data field,
    display min/max and color table info. Missing data get numpy.nan.
    Only called if rgb is True in write_grid_geotiff.

    Parameters
    ----------
    data : numpy.ndarray object, dtype int or float
        Two-dimensional data array.
    vmin : int or float
        Minimum value to color for RGB output or SLD file.
    vmax : int or float
        Maximum value to color for RGB output or SLD file.
    color_levels : int
        Number of color levels in cmap. Useful for categorical colormaps
        with steps << 255 (e.g., hydrometeor ID).
    cmap : str or matplotlib.colors.Colormap object, optional
        Colormap to use for RGB output or SLD file.

    Returns
    -------
    rarr : numpy.ndarray object, dtype int
        Red channel indices (range = 0-255).
    barr : numpy.ndarray object, dtype int
        Blue channel indices (range = 0-255).
    garr : numpy.ndarray object, dtype int
        Green channel indices (range = 0-255).

    """
    frac = (data - vmin) / np.float(vmax-vmin)
    if color_levels is None:
        color_levels = 255
    index = (frac * color_levels).ravel()
    # Out-of-bounds values will be lowest/highest colors
    index[index < 0] = 0
    index[index > 255] = 255
    rarr = []
    garr = []
    barr = []
    cmap = plt.cm.get_cmap(cmap)
    for val in index:
        if not np.isnan(val):
            ind = np.int(np.round(val))
            r, g, b, t = cmap(ind)
            rarr.append(np.int(np.round(r * 255)))
            garr.append(np.int(np.round(g * 255)))
            barr.append(np.int(np.round(b * 255)))
        else:
            rarr.append(np.nan)
            garr.append(np.nan)
            barr.append(np.nan)
    rarr = np.reshape(rarr, data.shape)
    garr = np.reshape(garr, data.shape)
    barr = np.reshape(barr, data.shape)
    return rarr, garr, barr


def _create_sld(cmap, vmin, vmax, filename, color_levels=None):
    """
    Develop a Style Layer Descriptor file given a color table and
    user-specified min/max files. Output color info to that file.
    Only called if sld is True in write_grid_geotiff.

    Parameters
    ----------
    cmap : str or matplotlib.colors.Colormap object, optional
        Colormap to use for RGB output or SLD file.
    vmin : int or float
        Minimum value to color for RGB output or SLD file.
    vmax : int or float
        Maximum value to color for RGB output or SLD file.
    filename : str
        Template for SLD filename. The suffix (presumably .tif or .tiff)
        is removed and replaced with .sld. Thus, if provided a filename
        radar_reflectivity.tif, the output SLD file will be called
        radar_reflectivity.sld.

    Other Parameters
    ----------------
    color_levels : int or None, optional
        Number of color levels in cmap. Useful for categorical colormaps
        with steps << 255 (e.g., hydrometeor ID).

    """
    cmap = plt.cm.get_cmap(cmap)
    if color_levels is None:
        color_levels = 255
    name, _ = filename.split('.')
    ofile = name + '.sld'
    fileobj = open(ofile, 'w')

    header = """<?xml version="1.0" encoding="UTF-8"?>
<sld:StyledLayerDescriptor xmlns="http://www.opengis.net/sld" xmlns:sld="http://www.opengis.net/sld" xmlns:ogc="http://www.opengis.net/ogc" xmlns:gml="http://www.opengis.net/gml" version="1.0.0">
    <sld:UserLayer>
        <sld:LayerFeatureConstraints>
            <sld:FeatureTypeConstraint/>
        </sld:LayerFeatureConstraints>
        <sld:UserStyle>
            <sld:Name>""" + str(name) + """</sld:Name>
            <sld:FeatureTypeStyle>
                <sld:Name>name</sld:Name>
                <sld:FeatureTypeName>Feature</sld:FeatureTypeName>
                <sld:Rule>
                    <sld:RasterSymbolizer>
                        <sld:ColorMap>"""
    fileobj.write(header)

    for i in np.arange(color_levels + 1):
        val = i * (vmax - vmin) / (color_levels) + vmin  # color_levels + 1
        rgbt = cmap(i)
        if i == 0:
            op = 0.0
        else:
            op = 1.0
        hexval = colors.rgb2hex(rgbt[0:3])
        wstr = '\n                            <sld:ColorMapEntry color=\"' + \
            str(hexval).upper() + '\" opacity=\"' + str(op) + \
            '\" quantity=\"' + str(val) + '\"/>'
        fileobj.write(wstr)

    footer = """
                        </sld:ColorMap>
                    </sld:RasterSymbolizer>
                </sld:Rule>
            </sld:FeatureTypeStyle>
        </sld:UserStyle>
    </sld:UserLayer>
</sld:StyledLayerDescriptor>
"""
    fileobj.write(footer)
    fileobj.close()
